#
# $Id: bstream.rb,v 1.12 2002/02/06 12:43:28 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   BlockStream Library
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################


#
#   String utility
######################################################################

def str_to_hash ( string )
  key = ""
  ret = { }
  string.split(/\r?\n/).each do |line|
    case line
    when /^!(\S+)\s+(.+)/ then
      ret[$1] = $2
    when /^!(\S+)$/ then
      key = $1
    else
      line << "\n"
      if ret.include?( key ) then
        ret[key] << line
      else
        ret[key] = line
      end
    end
  end
  return ret
end

def hash_to_str ( tbl )
  nl = "\n"
  lines = [ ]
  tbl.each_pair do |key, cont|
    cont = cont.to_s
    if contstr.index(nl) then
      lines << "!#{key}\n#{cont}"
    else
      lines << "!#{key} #{cont}"
    end
  end
  lines << ""
  return lines.join("\n")
end

def hash_to_pstr ( tbl )
  nl = "\n"
  lines = [ ]
  maxlen = 0
  tbl.each_key do |key|
    key = key.to_s
    maxlen = key.size if maxlen < key.size
  end
  tbl.each_pair do |key, cont|
    cont = cont.to_s
    if cont.index(nl) then
      lines << "!#{key}\n#{cont}"
    else
      lines << "!#{key.to_s.ljust(maxlen)} #{cont}"
    end
  end
  lines << ""
  return lines.join("\n")
end



#
#   String expansion
######################################################################

class String

  alias write concat

  def gets ( rs = $/ )
    return nil if self.empty?
   ($_ = self.index(rs))? self.slice!( 0, $_ + rs.length ): self.slice!( 0, self.length )
  end

  def display ( out )
    out << self
  end

  def to_hash ()
    str_to_hash( self )
  end

end


#
#   Array expansion
######################################################################

class Array

  alias write push

end


#
#   StringStream
######################################################################

class StringStream
  STR_SPLITTER = "\n"
  
  def initialize ( lines )
    if lines.kind_of?( Array ) then
      @lines = lines.dup
      @lines.collect! {|e| String(e) }
    else
      @lines = lines.to_s.dup
    end
    @lineno = 0
  end

  def gets ( rs = $/ )
    return nil  if @lines.empty?
    @lineno += 1
    return (Array === @lines)? @lines.shift: @lines.gets( rs )
  end
  
  def path ()
    "<(String)>"
  end
  
  def close ()
    @lines.clear()
  end
  
  def closed? ()
    @lines.empty?
  end
  
  alias eof  closed?
  alias eof? eof
  
  attr_accessor :lineno
end


#
#   Buffered Stream
######################################################################

BEGIN {
  require "nkf"
  if /^[Nn]/ === $KCODE then
    $NKFOPT = 128
  else
    $NKFOPT = nil
  end
}

class BufferedStream
  
  def initialize ( stream )
    @stream = stream
    @cache  = [ ]
    @lineno = stream.lineno
  end
  
  def gets ( rs = $/ )
    if @cache.empty? then
      $_ = @stream.gets( rs )
      if $_ and $NKFOPT then
        case NKF.guess( $_ )
        when NKF::EUC then
          $KCODE = "EUC"
          $NKFOPT = nil
        when NKF::SJIS then
          $KCODE = "SJIS"
          $NKFOPT = nil
        else
          $NKFOPT -= 1
          $NKFOPT = nil if $NKFOPT <= 0
        end
      end
    else
      $_ = @cache.shift()
    end
    @lineno += 1
    return $_
  end
  
  def ungets ( line )
    @cache << line
    @lineno -= 1
  end
  
  def readline ()
    $_ = gets()
    raise( EOFError, "reached EOF." ) unless $_
    return $_
  end
  
  def readlines ()
    lines = [ ]
    lines << $_ while $_ = gets()
    return lines
  end
  
  def eof ()
    @stream.eof()
  end
  
  alias eof? eof
  
  def close ()
    @stream.close()
  end
  
  def closed? ()
    @stream.closed?
  end
  
  def path ()
    if defined? @stream.path then
      @stream.path
    else
      case @stream
      when STDIN
        "<STDIN>"
      when STDOUT
        "<STDOUT>"
      when STDERR
        "<STDERR>"
      else
        nil
      end
    end
  end
  
  attr_accessor  :lineno
end


#
#   General Block
######################################################################

module Block
  RE_OPEN  = /^<([-_\w]*?)>$/
  RE_CLOSE = /^<\/([-_\w]*?)>$/
  class Monitor
    def initialize ( block )
      @block    = block
      @string   = [ ]
      @children = [ ]
    end
    def display ( out )
      n = @block.name
      raise "Block(#{n}) is not closed." unless @block.closed?
      out << "<#{n}>\n" unless n.empty?
      out << @string.join('')
      out << "\n" if !@string.empty? and @string.last[-1] != ?\n
      for $_ in @children do
        $_.display( out )
      end
      out << "</#{n}>\n" unless n.empty?
      self
    end
    def write ( line )
      @string << line
      self
    end
    def push ( child )
      @children << child
      self
    end
    def delete_child ( n = nil )
      if block_given? then
        @children.delete_if do |child|
          yield( child.block.name )
        end
      else
        @children.delete_if do |child|
          n === child.block.name
        end
      end
      self
    end
    def push_dummy ( blockname )
      @children << Monitor.new( DummyBlock.new( blockname.to_s ) )
      @children.last
    end
    attr_reader :block, :string, :children
  end
  
  def startline? ( line )
    (RE_OPEN === line)? $1: nil
  end
  def closeline? ( line )
    (RE_CLOSE === line)? (@tagname == $1): false
  end
  protected :startline?, :closeline?
  
  STR_SPLITTER = "\n"
  def block_init ( object )
    if PacketBlock === object then
      @parent = object.lock()
      @stream = object.stream
      
      @monitor = (@parent.monitor)? Monitor.new( self ): nil
      linestack = [ ]
      begin # ubN܂ŃXg[ǂݍ
        $_ = @stream.gets( STR_SPLITTER )
        linestack.push( $_ )
        unless $_ then
          @stream.ungets( linestack.shift() ) until linestack.empty?
          raise "Block is not found. (#{@stream.path}:#{@stream.lineno})"
        end
        @tagname = startline?( $_ )
      end until @tagname
    else
      @parent = nil
      @stream = BufferedStream.new( object ) unless BufferedStream === object
      
      @monitor = nil
      @tagname = ""
    end
    @locked = 0
    @closed = false
  end
  
  def capture ()
    @monitor = Monitor.new( self ) unless @monitor
    return @monitor
  end
  
  def level ()
    (@parent)? @parent.level + 1: 0
  end
  
  def locked? ()
    raise "Block(#{path}) is locked." if @locked > 0
  end
  protected :locked?
  
  def lock ()
    @locked += 1
    self
  end
  
  def unlock ()
    raise "Block(#{path}) is not locked." if @locked <= 0
    @locked -= 1
  end
  
  def path ()
    if @parent then
      (@parent.level > 0)? "#{@parent.path}/#{@tagname}": "/#{@tagname}"
    else
      '/'
    end
  end
  
  attr_reader :closed
  alias closed? closed
  
  def lineno ()
    @stream.lineno
  end
  
  attr_reader :monitor, :tagname, :stream
  alias name tagname
end


#
#   DataBlock
######################################################################

class DataBlock
  include Block
  
  RE_NAME = /^[Dd]/
  def initialize ( object )
    block_init( object )
    unless !@parent or RE_NAME === @tagname then
      raise "Cannot use `#{@tagname}' as name of DataBlock. (#{@stream.path}:#{@stream.lineno})"
    end
  end
  
  def close ()
    return if @closed
    locked?
    
    if @parent then
      begin
        until closeline?( $_ = @stream.gets( STR_SPLITTER ) ) do
          raise "Terminal line of Block(#{self.path}) is not found." unless $_
          @monitor.write( $_ ) if @monitor
        end
      ensure
        @parent.unlock()
      end
    else
      @stream.close() # e == [gpPbg
    end
    @closed = true
  end
  
  def gets ()
    return nil if @closed
    locked?
    
    $_ = @stream.gets( STR_SPLITTER )
    if closeline?( $_ ) then
      @stream.ungets( $_ )
      close()
      return nil
    end
    @monitor.write( $_ ) if @monitor
    
    return $_
  end
  
  def readlines ()
    return [] if @closed
    locked?
    lines = [ ]
    lines << $_ while $_ = gets()
    return lines
  end
  
  def read ()
    return [] if @closed
    locked?
    lines = [ ]
    lines << $_ while $_ = gets()
    return lines.join('')
  end
  
  def each_line ()
    locked?
    while $_ = gets() do
      yield( $_ )
    end
  end
  alias each each_line
  
  def each_child ()
    # q͐΂ɑ݂Ȃ̂ŃbN`FbN̂ݍs
    locked?
  end
  alias each_block each_child
end


#
#   PacketBlock
######################################################################

class PacketBlock
  include Block
  
  RE_NAME = /^[pP]/
  def initialize ( object )
    block_init( object )
    unless !@parent or RE_NAME === @tagname then
      raise "Cannot use `#{@tagname}' as name of PacketBlock. (#{@stream.path}:#{@stream.lineno})"
    end
  end
  
  def create_block ()
    line    = @stream.gets()
    tagname = startline?( line )
    return false unless tagname
    
    if DataBlock::RE_NAME === tagname then
      typeval = DataBlock
    elsif PacketBlock::RE_NAME === tagname then
      typeval = PacketBlock
    else
      typeval = nil
    end
    @stream.ungets( line )
    raise "Cannot use `#{tagname}' as name of Block. (#{@stream.path}:#{@stream.lineno})" unless typeval
    
    return typeval.new( self )
  end
  private :create_block
  
  def close ()
    return if @closed
    locked?
    
    if @parent then
      begin
        until closeline?( $_ = @stream.gets() ) do
          raise "Terminal line of Block(#{self.path}) is not found." unless $_
          @stream.ungets( $_ )
          
          if block = create_block() then
            @monitor.push( block.monitor ) if @monitor
            block.close()
          end
        end
      ensure
        @parent.unlock()
      end
    else
      @stream.close() # e == [gpPbg
    end
    @closed = true
  end
  
  def getb ()
    return nil if @closed
    locked?
    
    while $_ = @stream.gets() do
      @stream.ungets( $_ )
      if closeline?( $_ ) then # I[sɒH蒅I
        close()
        return nil
      end
      if block = create_block() then # ubNJns΂Ԃ
        @monitor.push( block.monitor ) if @monitor
        return block
      end
    end
    
    return nil
  end
  
  def each_child ()
    return if @closed
    locked?
    
    while $_ = getb() do
      yield( $_ )
      $_.close()
    end
  end
  alias each       each_child
  alias each_block each_child
  
  def each_line ()
    # f[^͐΂ɑ݂Ȃ̂ŃbN`FbN̂ݍs
    locked?
  end
end


#
#   Dummy Block
######################################################################

class DummyBlock
  def initialize ( blockname )
    @name = blockname.to_s.dup
  end
  
  def closed? ()
    true
  end
  
  attr_reader :name
end


#
#   Global Method
######################################################################

def open_packetfile ( input )
  # filename ̏ꍇ open 邪ȊȌꍇ͂̂܂ܒʂ
  input = open( input ) if String === input
  return PacketBlock.new( input ).getb()
end


#
#   Extention
######################################################################

require File.dirname(File.expand_path(__FILE__)) + "/bstream.so"
