#
# $Id: nlputil.rb,v 1.17 2002/02/08 12:00:39 Black Exp $
#                                                      Project Gemini
#---------------------------------------------------------------------
#
#   NLP utilities
#
#       Copyright (C) 2001-2002, ABE Shigeru <abe@mi.cs.titech.ac.jp>
######################################################################
require File.dirname(File.expand_path(__FILE__)) + "/bstream.rb"


#
#   Global Methods
######################################################################

RE_NULLLINE = /\A\s*\Z/
def nullline? ( line )
  RE_NULLLINE === line
end


#
#   Stream expansion
######################################################################

class IO
  def path ()
    case self
    when STDIN
      "<STDIN>"
    when STDOUT
      "<STDOUT>"
    when STDERR
      "<STDERR>"
    else
      "<IO>"
    end
  end
end


#
#   Struct::Tms expansion
######################################################################

class Struct
  class Tms
    def + ( tms )
      Struct::Tms.new( self.utime + tms.utime, self.stime + tms.stime, self.cutime + tms.cutime, self.cstime + tms.cstime )
    end
    def ptime ()
      self.utime + self.stime
    end
  end
end


#
#   Exception related NLP
######################################################################

class NLPException < RuntimeError
  def initialize ( *arg )
    super( *arg )
  end
end


#
#   Sequential ID module
######################################################################

module Sequential
  
  def id= ( i )
    @id = ((i)? i.to_i: nil)
  end
  
  attr_reader :id
  alias  to_i    id
  alias  to_int  id
  
end


#
#   Helper of reading a string
######################################################################

module StringParser
  
  def self.eat! ( re, s )
    if re === s then
      ($`.empty?)? s.replace( $' ): s.replace( $` + $' )
      Regexp.last_match
    else
      nil
    end
  end
  
  RE_category = /^\s*(<.+?>|[^\s<>(){}\[\]]+)/
  def self.eat_category! ( s )
    if RE_category === s then
      s.replace( $' )
      return $1
    else
      return nil
    end
  end
  
  RE_modifier = /^\s*\((.*?)\)/
  def self.eat_modifier! ( s )
    if RE_modifier === s then
      s.replace( $' )
      return $1
    else
      return nil
    end
  end
  
end


#
#   Grammar Object
######################################################################

class Grammar
  
  class Rule
    include Sequential
    
    class Category
      include Sequential
      
      def initialize ( string )
        @string = String( string )
        @id = nil
      end
      
      def nonterm? ()
        @string[0] == ?< and @string[-1] == ?>
      end
      
      def preterm? ()
        @string[0] != ?< or @string[-1] != ?>
      end
      
      def inspect ()
        @string + (@id)? "(\##{@id})": "(#?)"
      end
      
      def eql? ( r )
        @string == r.to_s
      end
      
      def hash ()
        @string.hash()
      end
      
      def to_s ()
        @string
      end
      
      alias == eql?
    end
    # class Category
    
    def initialize ( lhs, rhs )
      @lhs  = lhs
      @rhs  = Array( rhs )
      @mod  = { }
      @id   = nil
      @hash = (@rhs.dup << @lhs).hash() # E + ӂ̔zŃnbVvZ
    end
    
    def eql? ( r )
      @lhs == r.lhs and @rhs == r.rhs
    end
    alias == eql?
    
    attr_reader :hash
    
    def [] ( key )
      @mod[ key ]
    end
    
    def []= ( key, value )
      @mod[ key ] = value
    end
    
    def inspect ()
      "[\##{@id}: #{to_s}]"
    end
    
    def to_s ()
      if @mod.empty? then
        "#{@lhs} --> #{@rhs.join(' ')}"
      else
        m = [ ] ; @mod.each_pair {|k,v| m << ((v == true)? k: "#{k}=#{v}") }
        "(#{m.join(',')}) #{@lhs} --> #{@rhs.join(' ')}"
      end
    end
    
    def typestring ( n = nil )
      if n then
        n, rhs = Integer(n), @rhs.dup
        rhs[n,0] = '*'
      else
        rhs = @rhs
      end
      "#{@lhs} --> #{rhs.join(' ')}"
    end
    
    def freeze ()
      @rhs.freeze()
      super()
    end
    
    STR_pr = 'p'
    def pr= ( prob )
      @mod[ STR_pr ] = prob
    end
    def pr ()
      @mod[ STR_pr ]
    end
    
    attr_reader :lhs, :rhs, :mod
    alias top lhs
  end
  # class Rule
  
  def  renumbering  ()
    return if @nonterm.empty? or @preterm.empty?
    
    # nonterm  preterm ̋EA ID Ȃ΂ׂĐƂ
    unless @preterm.first.id == @nonterm.last.id + 1 then
      preterm_id = @nonterm.last.id + 1
      @preterm.each do |cat|
        cat.id = preterm_id
        preterm_id += 1
      end
    end
  end
  private :renumbering
  
  def store_cat ( catstr )
    return $_ if $_ = @catmap[ catstr ]
    
    cat = Rule::Category.new( catstr )
    if cat.nonterm? then
      cat.id = ((@nonterm.empty?)? 0: @nonterm.last.id+1)
      @nonterm << cat
    else
      cat.id = ((@preterm.empty?)? ((@nonterm.empty?)? 0: @nonterm.last.id+1): @preterm.last.id+1)
      @preterm << cat
    end
    unless cat.id then
      raise "Category-ID is not updated."
    end
    renumbering()
    @catary[ cat.id, 0 ] = cat
    
    @catmap[ catstr ] = cat
    return cat
  end
  private :store_cat
  
  RE_MODSplit  = /,/
  RE_UNITSplit = /=/
  def parse_modstr ( modstr )
    mod = { }
    modstr.split( RE_MODSplit ).each do |unit|
      $_ = unit.split( RE_UNITSplit, 2 )
      mod[ $_[0] ] = $_[1] || true
    end
    return mod
  end
  private :parse_modstr
  
  RE_RULESplit = /--+>/
  def parse_rule ( string )
    string = string.to_s.dup
    
    # 𒊏o
    modstr = StringParser.eat_modifier!( string )
    mod = (modstr)? parse_modstr( modstr ): nil
    
    # K𒊏o
    lhs, *rhs = Ext::StringParser::parse_rule( string )
    
    # Lϊ
    lhs = store_cat( lhs )
    rhs.map! {|cat| store_cat( cat ) }
    
    # @̐
    rule = Rule.new( lhs, rhs )
    rule.mod.replace( mod ) if mod
    
    return rule
  end
  private :parse_rule
  
  def store_rule ( rule )
    @rulemap[ rule ] = rule
    rule.id = @ruleary.length
    @ruleary << rule
    @rulestrmap[ rule.typestring ] = rule
    rule.freeze
  end
  private :store_rule
  
  RE_MULTI_SPACE = /\s\s+/
  STR_UNI_SPACE = ' '
  def str_to_rule ( string )
    return $_ if $_ = @rulestrmap[ string ]
    
    string = string.sub( RE_MULTI_SPACE, STR_UNI_SPACE )
    return $_ if $_ = @rulestrmap[ string ]
    
    rule = parse_rule( string )
    return $_ if $_ = @rulemap[ rule ]
    store_rule( rule )
  end
  alias [] str_to_rule
  
  RE_FIRSTSyntax = /^\s*([Ff]irst|FIRST)\s*\(\s*(.+?)\s*\)\s*=/
  def read_FIRST ( block )
    block.each do |line|
      next if nullline?( line )
      
      if RE_FIRSTSyntax === line then
        cat = $2
        list = $'.split(' ')
        list.compact!
        @FIRST[cat] = list
      else
        raise( NLPException, "FIRST syntax error." )
      end
    end
  end
  private :read_FIRST
  
  NAME = /^[pP][gG]rammar$/
  def self.proper? ( block )
    Ext::PacketBlock === block and NAME === block.name
  end
  
  RE_Rule  = /^[dD][rR]ules?$/
  RE_FIRST = /^[dD]([fF]irst|FIRST)$/
  def initialize ( block = nil )
    @rulemap    = { }
    @rulestrmap = { } # String --> Rule
    @ruleary    = [ nil ]
    
    @catmap  = { } # String --> Category
    @nonterm = [ ]
    @preterm = [ ]
    @catary  = [ ]
    
    @FIRST = nil
    
    # ubNw肳Ăꍇ
    if Grammar === block then
      block.each do |child|
        begin
          case child.name
          when RE_Rule then
            child.each_line do |line|
              rule = parse_rule( line )
              raise "Same rules are found. (#{rule.typestring})" if @rulemap.include?( rule )
              store_rule( rule )
            end
          when RE_FIRST then
            @FIRST = { }
            read_FIRST( child )
          end
        rescue NLPException
          raise "#{child.stream.path}:#{child.stream.lineno}: #{$!}"
        end
      end
      self.freeze()
    end
  end
  
  def freeze_FIRST ()
    return if !@FIRST or @FIRST.empty?
    _f = { }
    @FIRST.each_pair do |key, list|
      raise( NLPException, "Unknown category `#{key}' is found in FIRST set." ) unless @catmap.include?( key )
      list.map! do |cat|
        raise( NLPException,  "Unknown category `#{cat}' is found in FIRST(#{key})." ) unless $_ = @catmap[cat]
        $_
      end
      _f[ @catmap[key] ] = list
    end
    @FIRST = _f
    @FIRST.freeze()
  end
  
  def freeze ()
    renumbering() #unless self.frozen? #$$$#
    
    freeze_FIRST()
    @ruleary.freeze()
    @catary.freeze()
    for cat in @catary do
      cat.freeze()
    end
    
    super()
  end
  
  def size_rule    ()  @ruleary.size  end
  def size_nonterm ()  @nonterm.size  end
  def size_symbol  ()  @catary.size   end
  
  def to_a ()
    @ruleary.dup
  end
  alias rules to_a
  
  def getcat ( key )
    (Integer === key)? @catary[key]: @catmap[key.to_s]
  end
  
  def categories ()
    @catary.dup
  end
  
  def display ( out )
    Ext::Block.display( "pGrammar", out ) do
      # dRule ubN̏o
      Ext::Block.display( "dRules" ) do
        $defout << @ruleary[1..-1].join("\n") << "\n"
      end
      
      # dFIRST ubN̏o
      if @FIRST then
        Ext::Block.display( "dFIRST" ) do
          @FIRST.each_pair do |key, value|
            $defout << "FIRST( #{key} ) = #{value.join(' ')}\n"
          end
        end
      end
    end
  end
  
  attr_reader :FIRST
  
end
class << Grammar
  alias === proper?
end


#
#   PoS (terminal) list
######################################################################

class PoSList
  class LabelArray
    RNG_ZERO = 0..0
    def initialize ()
      @array = [ ]
      @label = nil
      @range = RNG_ZERO
    end
    def bracketed? ()
      (@label)? false: true
    end
    def tree? ()
      (@label)? true: false
    end
    def length ()
      @range.last - @range.first
    end
    def first ()
      @range.first
    end
    def last ()
      @range.last
    end
    attr_reader :array
    alias to_a array
    attr_accessor :range, :label
  end
  # class LabelArray
  
  Topen  = '[' ; Tclose = ']'
  Bopen  = '(' ; Bclose = ')'
  Lopen  = '{' ; Lclose = '}'
  def  parse_tree  ( strary, left )
    # '[' ΍\
    return nil unless Topen === strary.first
    strary.shift
    
    # gbṽx𓾂
    elem = LabelArray.new()
    elem.label = strary.shift
    
    # q\zĂ
    ary = elem.array
    right = left
    until Tclose === strary.first do
      if $_ = parse_tree( strary, right ) then
        @list << $_
        ary << -(@list.length - 1)
        right = @list.last.range.last
      else
        raise( NLPException, "[ and ] are not balanced." ) if strary.empty?
        ary << strary.shift
        @catlist << ary.last
        right += 1
      end
    end
    strary.shift
    elem.range = left .. right
    return elem
  end
  def  parse_termlist  ( strary )
    ret = [ ]
    until Lclose === strary.first do
      raise( NLPException, "{ and } are not balanced." ) if strary.empty?
      ret << strary.shift
    end
    strary.shift
    return ret
  end
  def  parse_bracket  ( strary, left )
    elem = LabelArray.new()
    ary = elem.array
    right = left
    until strary.empty? or Bclose === strary.first do
      if Bopen === strary.first then
        strary.shift
        @list << parse_bracket( strary, right )
        ary << -(@list.length - 1)
        right = @list.last.range.last
      else
        if Lopen === strary.first then
          strary.shift
          ary << parse_termlist( strary )
        else
          ary << strary.shift
        end
        @catlist << ary.last
        right += 1
      end
    end
    strary.shift
    elem.range = left .. right
    return elem
  end
  private :parse_tree, :parse_termlist, :parse_bracket
  
  RE_Splitter = /\s+|([()\[\]{}])/
  def initialize ( line )
    @compiled = false
    strary = line.split(RE_Splitter).delete_if {|e| !e or e.empty? }
    begin
      @catlist = [ ]
      @list = [ nil ] # Ō [0] Ŋi[邽ߋ ([]) Ƃ܂
      @list[0] = parse_tree( strary, 0 ) || parse_bracket( strary, 0 )
      raise( NLPException, "Parsing failed." ) unless strary.empty?
    rescue NLPException
      raise( NLPException, "#{$!} (#{line.strip})" )
    end
  end
  
  def compile ( grammar = $grammar )
    return self if @compiled
    
    getcat = lambda { |val|
      if String === val then
        cat = grammar.getcat( val )
        raise( NLPException, "PoS `#{val}' is unknown." ) unless cat
        val = cat
      end
      val.to_i
    }
    
    localcompile = lambda { |list|
      list.map! do |s|
        if Array === s then
          s.map! {|c| getcat[c] }
          s
        else
          getcat[s]
        end
      end
    }
    
    @list.each do |line|
      localcompile[ line.array ]
      line.label = getcat[ line.label ]
    end
    localcompile[ @catlist ]
    
    @compiled = true
    self
  end
  
  def at ( n )
    n = n.abs
    (n >= @list.length)? nil: @list.at( n )
  end
  
  def  length  ()
    @list.first.length
  end
  
  attr_reader :list, :catlist
  alias to_a list
  alias []   at
  
  attr_reader :compiled
  alias compiled? compiled
end


#
#   Corpus Components
######################################################################

class SentenceBlock
  NAME = /^[pP][sS]entence$/
  def self.proper? ( block )
    Ext::PacketBlock === block and NAME === block.name
  end
  
  def initialize ( block )
    unless SentenceBlock.proper?( block ) then
      raise( NLPException, "Failed in SentenceBlock initialization." )
    end
    @block = block
  end
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close()
    self
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def display ( out = $defout, &procedure )
    @block.display( out, &procedure )
  end
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
  
  def each_element ( *list )
    list.flatten!
    for i in list do
      raise "SentenceBlock.each_element(): An element of list must be Class." unless Class === i
    end
    @block.each do |block|
      for i in list do
        next unless i.proper?( block )
        yield( i, block )
        block = nil
        break
      end
      if block then
        yield( nil, block )
      end
    end
    return self
  end
  alias each each_element
end
class << SentenceBlock
  alias === proper?
end


#
#   ItemSet
######################################################################

class ItemSet
  class Item
    def initialize ( rule, posary )
      @rule   = rule
      @posary = posary
      @range  = posary.first .. posary.last
    end
    def top ()
      @rule.top
    end
    def hash ()
      @rule.hash
    end
    def eql? ( target )
      @rule == target.rule and @posary == target.posary
    end
    def left ()
      @range.first
    end
    def right ()
      @range.last
    end
    def to_s ()
      "[#{@posary.join(',')}] #{@rule.typestring}"
    end
    def inspect ()
      "<Item:[#{@posary.join(',')}] #{@rule.inspect}>"
    end
    def compile ( grammar )
      @rule = grammar.str_to_rule( @rule ) if String === @rule
    end
    attr_accessor :rule
    attr_reader :posary, :range
    alias to_a posary
    alias to_r rule
  end
  
  RE_comma  = /,/
  RE_posary = /^\s*\[(.+?)\]\s+/
  RE_digit  = /^\d+$/
  
  def parse_line ( line )
    posary = [ ]
    rule = Ext::StringParser::parse_item_poslist( line, posary )
    raise( NLPException, "Position list is not found." ) unless rule
    raise( NLPException, "Definition no.#{$_} is out of range." ) if Fixnum === rule and rule >= @rulemap.size
    
    return Item.new( rule, posary )
  end
  private :parse_line
  
  def compile ( grammar = $grammar )
    return self if @compiled
    
    @rulemap.collect! do |rule|
      (rule)? grammar.str_to_rule( rule ): nil
    end
    @rulemap.freeze
    
    starts   = [ ]
    rightpos = 0
    itemlist.each do |item|
      item.rule = @rulemap[ item.rule ]
      unless item.posary.size == item.rule.rhs.size + 1 then
        raise( NLPException, "Position list is short or long." )
      end
      
      # Jn̂߂Ƀf[^W
      rightpos = item.right if item.right > rightpos
      starts << item if item.top == @startname
    end
    
    # Jn
    @startary = [ ]
    starts.each do |item|
      @startary << item if item.right == rightpos
    end
    raise( NLPException, "Cannot find top items." ) if starts.empty?
    
    @compiled = true
    self
  end
  
  RE_root    = /^\#@ROOT\s*=\s*/
  RE_ruledef = /^\#(\d+)\s+/
  def initialize ( lines )
    @itemlist  = [ ]
    @startary  = nil
    @startname = $StartName
    @rulemap   = [ ]
    rootname   = nil
    
    # es獀擾 @itemlist ɕ荞
    n = 0
    for line in lines do
      case line
      when RE_root then
        raise( NLPException, "Found too many \#@ROOT direction." ) if rootname
        @startname = rootname = $'.strip
      when RE_NULLLINE then
        ; # Ȃ
      when RE_ruledef then
        @rulemap[ $1.to_i ] = $'
      else
        @itemlist << parse_line( line )
      end
    end
    
    @compiled = false
  end
  
  attr_reader :startary, :itemlist, :rulemap
  alias to_a itemlist
  
  attr_reader :compiled
  alias compiled? compiled
end


#
#   Support Graph
######################################################################

class SupportGraph
  
  def initialize ( lines )
    @count_node = @count_path = 0
    @sublist = [ ]
    @comment = [ ]
    
    subgraph = graphnum = nil
    lines.each do |line|
      if $_ = Ext::StringParser::sgraph_path?( line ) then
        raise( NLPException, "Syntax error. (uninitialized subgraph)" ) unless subgraph
        @count_node += $_[0]
        @count_path += 1
        subgraph << $_[1]
      elsif $_ = Ext::StringParser::sgraph_index?( line ) then
        graphnum, comment = $_
        
        # ̕@Ɓuv󂭂Ȃ̂Œ  RpCɌo
        @comment[ graphnum ] = comment
        subgraph = @sublist[ graphnum ]
        subgraph = @sublist[ graphnum ] = [ ] unless subgraph
      end
    end
    @compiled = false
  end
  
  def compile ( database = $grammar )
    @sublist.each do |subgraph|
      subgraph.each do |pathdat|
        pathdat.map! do |node|
          if String === node then
            $_ = database[ node ]
            unless $_ then
              raise( NLPException, "Cannot resolve a switch node `#{node}'." )
            end
            $_
          else
            node
          end
        end
      end
    end
    @compiled = true
    self
  end
  
  attr_reader :count_node, :count_path
  def count_graph
    @sublist.size
  end
  
  attr_reader :sublist, :comment
  alias to_a sublist
  
  attr_reader :compiled
  alias compiled? compiled
end


#
#   Support graph Block
######################################################################

class SupportGraphBlock
  NAME = /^[dD](Support[gG]raph|supportgraph)$/
  def self.proper? ( block )
    Ext::DataBlock === block and NAME === block.name
  end
  
  def initialize ( block )
    unless SupportGraphBlock.proper?( block ) then
      raise( NLPException, "Failed in SupportGraphBlock initialization." )
    end
    @block = block
  end
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close() if @block
    self
  end
  def display ( out = $defout, &procedure )
    @block.display( out, &procedure )
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
  
  def read ()
    return @sg if @sg
    
    raise( NLPException, "<dSupportGraph> is already closed." ) if @block.closed?
    @sg = SupportGraph.new( @block )
    @block.close()
    @block = nil
    return @sg
  end
end
class << SupportGraphBlock
  alias === proper?
end


#
#   Item set Block
######################################################################

class ItemSetBlock
  NAME = /^[dD](Item[sS]et|itemset)$/
  def self.proper? ( block )
    Ext::DataBlock === block and NAME === block.name
  end
  
  def initialize ( block )
    unless ItemSetBlock.proper?( block ) then
      raise( NLPException, "Failed in ItemSetBlock initialization." )
    end
    @block = block
  end
  def read ()
    return @itemset if @itemset
    
    raise( NLPException, "<dItemSet> is already closed." ) if @block.closed?
    @itemset = ItemSet.new( @block )
    @block.close()
    @block = nil
    return @itemset
  end
  def compile ( grammar = $grammar )
    read() unless @itemset
    @itemset.compile( grammar )
  end
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close() if @block
    self
  end
  def display ( out = $defout, &procedure )
    @block.display( out, &procedure )
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
end
class << ItemSetBlock
  alias === proper?
end


#
#   PoS Block
######################################################################

class PoSBlock
  NAME = /^[dD](PoS|pos)$/
  def self.proper? ( block )
    Ext::DataBlock === block and NAME === block.name
  end
  
  def initialize ( block )
    unless PoSBlock.proper?( block ) then
      raise( NLPException, "Failed in PoSBlock initialization." )
    end
    @block = block
  end
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close()
    self
  end
  def display ( out = $defout, &procedure )
    @block.display( out, &procedure )
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
  
  def read ()
    return @poslist if @poslist
    
    raise( NLPException, "<dPoS> is already closed." ) if @block.closed?
    line = @block.read()
    
    @poslist = PoSList.new( line )
    @block.close() ; @block = nil
    (block_given?)? yield( @poslist ) : @poslist
  end
  def compile ( grammar = $grammar )
    read() unless @poslist
    @poslist.compile( grammar )
  end
end
class << PoSBlock
  alias === proper?
end


#
#   Corpus
######################################################################

class CorpusBlock
  NAME = /^[pP][cC]orpus$/
  def self.proper? ( block )
    Ext::PacketBlock === block and NAME === block.name
  end
  
  def initialize ( block )
    unless CorpusBlock.proper?( block ) then
      raise( NLPException, "Failed in CorpusBlock initialization." )
    end
    @block = block
  end
  
  def get_sentence ()
    while b = @block.getb() do
      return SentenceBlock.new( b ) if SentenceBlock.proper?( b )
      b.close()
    end
    return nil
  end
  alias getb get_sentence
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close()
    return self
  end
  def display ( out = $defout, &procedure )
    @block.display( out, &procedure )
  end
  def dump ( out = $defout )
    @block.dump( out )
  end
  def out= ( out )
    @block.out = out
  end
  alias monitor= out=
  
  def each_sentence ()
    while s = get_sentence() do
      yield( s )
      s.close()
    end
  end
  alias each each_sentence
end
class << CorpusBlock
  alias === proper?
end


#
#   Corpus File Block
######################################################################

class CorpusStream
  def initialize ( stream )
    @block = Ext::Block.open( stream )
    if block_given? then
      yield( self )
      close()
    end
  end
  
  def get_corpus ()
    while b = @block.getb() do
      return CorpusBlock.new( b ) if CorpusBlock.proper?( b )
      b.close()
    end
    return nil
  end
  alias getb get_corpus
  
  def stream ()
    @block.stream
  end
  def close ()
    @block.close()
    return self
  end
  def each_corpus ()
    while c = get_corpus() do
      yield( c )
      c.close()
    end
    return self
  end
  alias each each_corpus
end


#
#   Utility
######################################################################


def v_print ( arg = nil )
  return unless $OPT_v
  $stderr.print( block_given? ? yield : arg )
end

def verbose_print ( arg = nil )
  return unless $OPT_verbose
  $stderr.print( block_given? ? yield : arg )
end


BEGIN {
  $get_process_time = 
    case RUBY_VERSION
    when /^1\.6\./
      lambda { Time::times }
    else #when /^1\.7\./
      lambda { Process::times }
    end
}
def get_process_time ()
  $get_process_time.call()
end


def time_block ( &proc )
  t1 = $get_process_time.call()
  $_ = proc.call()
  t2 = $get_process_time.call()
  
  $time = Struct::Tms.new( t2.utime - t1.utime, t2.stime - t1.stime, t2.cutime - t1.cutime, t2.cstime - t1.cstime )
  
  return $_
end

# ܂gpȂ (gp}sȌۂ)
def time_block_withoutGC ( &proc )
  GC.start
  GC.disable
  $_ = time_block( &proc )
  GC.enable
  
  return $_
end


def printVersion ()
  $stderr.print $VERSION
  exit
end

def printUsage ()
  $stderr.print $VERSION
  $stderr.print $USAGE
  exit
end

def stream_read_grammar ( stream )
  root = Ext::Block.open( stream )
  root.each do |block|
    if Grammar === block then
      begin
        grammar = Grammar.new( block )
        block.close()
      rescue NLPException
        stream = root.stream
        raise( NLPException, "#{root.stream.path}:#{root.stream.lineno}: #{$!}" )
      end
      root.close()
      return grammar
    end
  end
  return nil
end

require "getopts.rb"
def getopts_ex ( singleopt, *longopt )
  getopts( 'vh' + singleopt.to_s, 'help', 'verbose', 'version', 'g:', *longopt )
  
  $OPT_v = true if $OPT_verbose
  $OPT_h = true if $OPT_help
  
  if $OPT_g then
    begin
      unless File.exist?( $OPT_g ) then
        $stderr << "ERROR: No such grammar file -- #{$OPT_g}\n"
        exit
      end
      v_print { "now reading #{$OPT_g}\r" }
      $grammar = stream_read_grammar( open( $OPT_g ) )
      v_print { "            #{' ' * $OPT_g.length}\r" }
    rescue
      $stderr << $! << "\n"
      exit
    end
  end
  
  printVersion() if $OPT_version
  printUsage() if $OPT_h or $OPT_help
end

def io_to_name ( io )
  case io
  when STDIN
    "<STDIN>"
  when STDOUT
    "<STDOUT>"
  when STDERR
    "<STDERR>"
  else
    case io
    when File
      io.path
    else
      io.to_s
    end
  end
end


def toPCFG ( grammar, force_random = false )
  if !grammar.isPCFG? or force_random then
    if force_random then
      v_print { " + FORCE-mode is enabled.\n" }
    else
      v_print { " + Using grammar is not PCFG. (#{$!})\n" }
    end
    grammar.setRandomProbability()
    v_print { "    - Random probability is setted.\n" }
  end
  return grammar
end


#
#   Extention
######################################################################

require File.dirname(File.expand_path(__FILE__)) + "/nlputil.so"
